from .check import Check
import redis
import os
from redis import sentinel
import copy
# import pykafka
from kafka import KafkaProducer, KafkaConsumer

"""
# relational database centralized connection

"""
import time
import traceback

from contextlib import contextmanager
from sqlalchemy import create_engine, MetaData, Table
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool

import ostrich_json as json


class RDBC(object):
    def __init__(self, url, logger=None):
        """
        Do sqlalchemy initial work

        pool_pre_ping : boolean , ping the connection before use
        Session : db session instance

        """

        engine = create_engine(url, echo=False, poolclass=NullPool, pool_recycle=3600, pool_pre_ping=True)

        self.__metadata = MetaData(engine)

        self.__Session = sessionmaker()

        self.__Session.configure(bind=engine)

        self.__error_delay = 10

        self.logger = logger

    def table(self, table_name):
        """Auto load table structure , such as column name and table name"""

        user_table = Table(table_name, self.__metadata, autoload=True)
        return user_table

    @contextmanager
    def session(self):
        """Provide a transactional scope around a series of operations."""

        session = self.__Session()
        try:
            yield session
            session.commit()
        except (OperationalError, Exception) as E:
            if self.logger:
                self.logger.error(traceback.format_exc())
            session.rollback()
            time.sleep(self.__error_delay)
            raise
        finally:
            session.close()

    def query_all(self, mysql_table, expression=None):
        with self.session() as session:
            if expression:
                results = session.query(mysql_table).filter(eval(expression)).all()
            else:
                results = session.query(mysql_table).all()
            return results

    def delete_by_id(self, mysql_table, qid):
        with self.session() as session:
            instance = mysql_table.delete().where(mysql_table.c.id == qid)
            session.execute(instance)


class Redis(object):
    def __init__(self, redis_url, sequence=True, logger=None):
        if os.environ.get('CURRENT_ENV') != 'production':
            self.__redis_connection = redis.from_url(url=redis_url, socket_timeout=10, socket_connect_timeout=2)
        else:
            #  临时的集群连接参数
            master_name = os.environ.get('REDIS_SENTINEL_MASTER')
            startup_nodess = os.environ.get('REDIS_SENTINEL_NODES').split(',')
            startup_nodes = []
            for i in startup_nodess:
                startup_nodes.append((i.split(':')[0], int(i.split(':')[1])))
            password = os.environ.get('REDIS_PASSWORD')

            rc__ = sentinel.Sentinel(startup_nodes, socket_timeout=0.1)
            self.__redis_connection = rc__.master_for(master_name, password=password, socket_timeout=0.1, db=1)

        self.sequence = sequence
        self.logger = logger

    def origin(self):
        return self.__redis_connection

    def lpop(self, key):
        data = self.__redis_connection.lpop(key)
        return data

    def rpush(self, key, value):
        if value:
            return self.__redis_connection.rpush(key, value)

    def sismember(self, key, value):
        return self.__redis_connection.sismember(key, value)

    def sadd(self, key, value):
        return self.__redis_connection.sadd(key, value)

    def srem(self, key, value):
        return self.__redis_connection.srem(key, value)

    def srandmember(self, key):
        return self.__redis_connection.srandmember(key)

    def hset(self, key, field, value):
        return self.__redis_connection.hset(key, field, value)

    def json_pop(self, key, all=False):
        while 1:
            if self.__redis_connection.exists(key):
                if not all:
                    data = self.__redis_connection.lpop(key)
                    if json.is_json(data):
                        data = json.loads(data)
                        return data
                    elif not data:
                        if self.logger:
                            self.logger.debug('no data')
                        time.sleep(2)
                    else:
                        if self.logger:
                            self.logger.warn('not json data %s' % data)
                else:
                    data_all = []
                    while 1:
                        data = self.__redis_connection.lpop(key)
                        if json.is_json(data):
                            data = json.loads(data)
                            data_all.append(data)
                        elif not data:
                            if self.logger:
                                self.logger.debug('no data')
                            return data_all
                        else:
                            if self.logger:
                                self.logger.warn('not json data %s' % data)
            else:
                if self.logger:
                    self.logger.debug('no data')
                time.sleep(2)

    def json_push(self, key, value):
        if value:
            value = json.dumps(value)
            if self.sequence:
                self.__redis_connection.rpush(key, value)
            else:
                self.__redis_connection.lpush(key, value)


class Kafka(object):
    def __init__(self, broker):
        self.broker = broker

    def producer(self):
        kafka_producer = KafkaProducer(bootstrap_servers=self.broker)
        return kafka_producer

    def consumer(self, topic):
        kcm = KafkaConsumer(topic, bootstrap_servers=self.broker)
        return kcm


# class PyKafka(object):
#     def __init__(self, broker):
#         self.client = pykafka.KafkaClient(hosts=broker)
#
#     def producer(self, topic=''):
#         return self.client.topics[topic.encode('utf-8')].get_producer()
#
#     def produce(self, data, topic=''):
#         data = copy.deepcopy(data)
#         self.producer(topic=topic).produce(json.dumps(data).encode('utf-8'))
#
#     def consumer(self, topic, zookeeper_connect):
#         cosu = self.client.topics[topic.encode('utf-8')].get_balanced_consumer(
#             consumer_group=topic.encode('utf-8'), auto_commit_enable=True,
#             zookeeper_connect=zookeeper_connect)
#         return cosu


class ConnectionManager(object):
    def __init__(self, logger=None):
        self.__connections = dict()
        self.logger = logger

    def get(self, db_type, name, url):
        try:
            connection_instance = self.__connections.get(name)
            if connection_instance:
                connection_instance = connection_instance.get('instance')
            if connection_instance:
                return connection_instance
            else:
                return self.__set(db_type, name, url)
        except AttributeError as E:
            return self.__set(db_type, name, url)

    def __set(self, db_type, name, url):
        if db_type == 'redis':
            self.__connections[name] = {'instance': Redis(url), 'db_type': db_type}
        elif db_type == 'mysql':
            self.__connections[name] = {'instance': RDBC(url), 'db_type': db_type}
        elif db_type == 'kafka':
            self.__connections[name] = {'instance': Kafka(url), 'db_type': db_type}
        else:
            if self.logger:
                self.logger.error('%s connection not supported' % db_type)
            raise TypeError
        return self.__connections.get(name).get('instance')
